/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 ***************************************************************************
 * MODULE: Print_support.c
 * handle the actual output to the Printer
 ***************************************************************************
 * Revision History: Created Wed Jan 13 16:21:39 CST 1988
 * $Log:	print_support.c,v $
 * Revision 3.1  88/06/18  09:35:18  papowell
 * Version 3.0- Distributed Sat Jun 18 1988
 * 
 * Revision 2.4  88/05/27  08:28:01  papowell
 * Added a SIGCONT to start up the output filter.
 * 
 * Revision 2.3  88/05/19  10:34:19  papowell
 * Fixed open() calls to have a 0 parameter, ie: open(f, perms, 0), where needed
 * 
 * Revision 2.2  88/05/14  10:21:16  papowell
 * Modified -X flag handling
 * 
 * Revision 2.1  88/05/09  10:09:44  papowell
 * PLP: Released Version
 * 
 * Revision 1.4  88/03/25  15:01:04  papowell
 * Debugged Version:
 * 1. Added the PLP control file first transfer
 * 2. Checks for MX during file transfers
 * 3. Found and fixed a mysterious bug involving the SYSLOG facilities;
 * 	apparently they open files and then assume that they will stay
 * 	open.
 * 4. Made sure that stdin, stdout, stderr was available at all times.
 * 
 * Revision 1.3  88/03/11  19:29:04  papowell
 * Minor Changes, Updates
 * 
 * Revision 1.2  88/03/05  15:01:11  papowell
 * Minor Corrections,  Lint Problems
 * 
 * Revision 1.1  88/03/01  11:09:01  papowell
 * Initial revision
 * 
 ***************************************************************************
 * void Print_open(): opens the Printer
 * static int Print_of_fd(): makes an 'of' filter if needed
 * void Print_close(): closes the Printer
 * int Print_ready(): combines Print_open() and Print_of_fd()
 * static int of_stop(): stops the 'of' filter
 * int Print_string( str ): prints a string through 'of' or to Printer
 * int Print_copy( fd ): copies a file through 'of' or to Printer
 * int Print_filter( file, cmd ): makes a filter and copies file
 * int Print_banner(): prints a banner through 'of' or to Printer
 * Note: the above procedures which return values return JSUCC on success,
 *   and JFAIL or JABORT on failure
 ***************************************************************************/

#ifndef lint
static char id_str1[] =
	"$Header: print_support.c,v 3.1 88/06/18 09:35:18 papowell Exp $ PLP Copyright 1988 Patrick Powell";
#endif lint

#include "lp.h"

/***************************************************************************
 * Print_open()
 * 	Open the Printer, and set the Print_fd variables
 * 	If the RW printcap flag is set, output is opened RW, otherwise
 * 	opened writeonly in append mode.
 * 
 * 	If the Printer is a tty (i.e.- isatty returns non-zero),
 * 	then the baud rate is set.  The baud rate table is stolen from
 * 	the original LPD code, which stole it from STTY, etc.
 * Side Effect:
 * 	sets the Print_fd variable to a non-zero value
 *  terminates if Printer is unable to be opened
 ****************************************************************************/

struct bauds {
	char *string;
	int	baud;
	int	speed;
} bauds[] = {
	"50", 50,	B50,
	"75", 75,	B75,
	"110", 110,	B110,
	"134", 134,	B134,
	"150", 150,	B150,
	"200", 200,	B200,
	"300", 300,	B300,
	"600", 600,	B600,
	"1200", 1200,	B1200,
	"1800", 1800,	B1800,
	"2400", 2400,	B2400,
	"4800", 4800,	B4800,
	"9600", 9600,	B9600,
	"19200", 19200,	B19200,
	"38400", 38400,	B38400,
	(char *)0, 0,	0
};

static int of_pid;		/* OF process pid */
static int of_fd;		/* pipe to OF process */
static int of_running;	/* of filter running */
static char filter_stop[] = "\031\001";	/* what to send filter */

/***** ANNEX MODIFICATIONS ******/
#include <setjmp.h>
static jmp_buf ack_jmp_buf;
/*
 * SIGALRM catcher for Annex ACK listener.
 * If a SIGALRM comes, then we didn't get an initial ACK, meaning
 * that this Annex is probably using the old Annex LPD protocol.
 */
noannexack()
{
  longjmp(ack_jmp_buf, 1);
}

/*
 * Open a socket to an Annex terminal server.
 * This code has been excerpted from modifications made to the
 * Berkeley LPD code, by ENCORE Computers Inc., and has been used
 * with their permission.
 *
 * It appears that the terminal server accepts a port connection on the
 * LPD port.  I suppose that it then treats the selected printer nicely.
 * Not bad,  as far as hacks go,  but it does expose a severe security
 * loophole.
 *
 * Fri Sep 30 09:59:59 CDT 1988 Patrick Powell
 */
static int
Open_annex()
{
#define SERIAL  1
#define PAR   2
#define ANNEX_SPOOL   '\011'
#define ACK_WAIT 10   /* seconds to wait for annex to ACK */
	int i;			/* ACME Screw and Integers, Inc. even used by ENCORE */
	SIG_F_TYPE func;	/* what SIGALRM was set to */
	int pfd;		/* annex socket file descriptor */
	char line[BUFSIZ];	/* temporary buffer */
	int n;		/* number chars read/written */
	int err;		/* save value of errno */

	setstatus("trying to open Annex '%s' port '%d' since %s (offline?)",
		RA, AL, Time_str());
	(void) alarm(0); /* disarm */
	func = signal(SIGALRM, noannexack); /* set up time out function */
	for (i = 1; ; i = i < 512 ? i << 1 : i) {
		if(Debug>3)log(XLOG_DEBUG,
			"Open_annex: server '%s', line %d, port %d, delay %d",
			RA, AL, PO, i );
		pfd = getport(RA, PO );
		if (pfd >= 0) {
			/* see if you are simply connecting to a port */
			if( PO != 0 ){
				break;
			}
			/* you are trying to connect to the ANNEX PRINTER port,
			 * and set up a simple transfer protocol.
			 */
			(void) sprintf(line, "%c%d;%d\n",
				ANNEX_SPOOL,
				AL <= 0 ? PAR : SERIAL,
				AL <= 0 ? 1 : AL);
			n = strlen(line);
			if(Debug>3)log(XLOG_DEBUG,"Open_annex: sending '%d''%s'",
				line[0], line+1);
			/*
			 * make sure that it is set correctly, sleep in some
			 * implementations will clobber the SIGALRM value and not
			 * restore it
			 */
			(void)signal(SIGALRM, noannexack); /* set up time out function */
			if(setjmp(ack_jmp_buf)) {
				logerr( XLOG_INFO,
					"Annex '%s', port '%d' response timed out, old ANNEX?",
					RA, AL );
			} else {
				(void) alarm(ACK_WAIT);
				if (write(pfd, line, n) == n) {
					if(Debug>3)log(XLOG_DEBUG,
						"Open_annex: waiting for response");
					n = read( pfd, line, 1 );
					err = errno;
					(void) alarm(0); /* disarm */
					errno = err;
					if( n != 1 ){
						logerr( XLOG_INFO,
							"Open_annex: read error '%s', port '%d'", RA, AL);
					} else if( line[0] != '\0' ){
						log( XLOG_INFO,
							"Open_annex: '%d' read from '%s', port '%d'",
							line[0],RA, AL);
					} else {
						/* success */
						/* this should exit from the loop */
						break;
					}
				} else {
					(void) alarm(0); /* disarm */
					if(Debug>3)logerr(XLOG_DEBUG,
						"Open_annex: write failed '%s',port '%d'", RA, AL );
				}
			}
			(void) close(pfd);
		} else {
			logerr( XLOG_INFO,
				"Open_annex: open error '%s', port '%d'", RA, AL);
		}
		/*
		 * note that sleep may clobber SIGALMR entry
		 */
		sleep((unsigned)i);
	}
	(void) alarm(0); /* disarm */
	(void)signal(SIGALRM, func);
	setstatus("sending to Annex '%s' line/port '%d' at %s", RA,
		PO?PO:AL, Time_str());
	if(Debug>3)log(XLOG_DEBUG,"Open_annex: sending to '%s' port '%d'", RA,
		PO?PO:AL );
	return( pfd );
}
/********************************/

Print_open()
{
	int err;

	/*
	 * check to see if it is open
	 */
	if( Print_fd ){
		return;
	}
	if( (LP == 0 || *LP == 0) && (RA == 0 || *RA == 0) ){
		fatal( XLOG_INFO, "Missing both LP and RA value for local printer");
	}
	if( LP ){
		setstatus("waiting for %s to become ready since %s (offline?)",
			Printer, Time_str());
		Print_fd = open(LP,
			RW ? ( O_RDWR|O_APPEND ) : ( O_WRONLY|O_APPEND ), 0);
	} else {
		Print_fd = Open_annex();
	}
	if( Print_fd < 0 ){
		err = errno;
		setstatus("Print_open: '%s' open(%s) failed, %s",Printer,LP?LP:RA,
			Errormsg(err));
		errno = err;
		logerr_die( XLOG_INFO, "Print_open: cannot open %s", LP?LP:RA);
	} else if( Print_fd == 0 ){
		fatal( XLOG_INFO, "Print_open: cannot happen- Print_fd == 0");
	}
	/*
	 * if it is a tty, set the baud rates and control information
	 */
	if (isatty(Print_fd)){
		Do_stty(Print_fd);
	}
	setstatus("printer %s output %s opened at %s",
		Printer, LP?LP:RA, Time_str());
	if(Debug>3)log(XLOG_DEBUG,"Print_open: %s is fd %d", LP?LP:RA, Print_fd );
}

/*
 * Start up an output filter, if needed.
 */
Print_of_fd()
{
	int p[2];
	char *cmd;

	if( OF == 0 || *OF == 0 || of_pid ){
		return;
	}
	/*
	 * set OF command line
	 */
	cmd = Setup_filter('o',OF);
	if( cmd == 0 || *cmd == 0 ){
		fatal(XLOG_INFO, "bad OF entry");
	}
	if(Debug>2)log( XLOG_DEBUG, "starting OF, '%s'", cmd);
	/*
	 * create pipe and fork
	 */
	if( pipe(p) < 0 ){
		logerr_die( XLOG_NOTICE, "Print_of_of: pipe failed" );
	}
	if( p[0] < 3 || p[1] < 3 ){
		fatal( XLOG_INFO, "Print_of_fd: IMPOSSIBLE- pipe fd %d %d", p[0], p[1]);
	}
	if ((of_pid = fork()) == 0) {	/* child */
		if( dup2(p[0], 0) < 0		/* pipe is std in */
		 || dup2(Print_fd, 1) < 0 ){		/* Printer is std out */
			logerr_die( XLOG_NOTICE, "Print_of_fd: dup2 failed" );
		}
		if(Debug>3)log( XLOG_DEBUG, "Print_of_fd: forked %d OF: %s",
			getpid(), cmd );
		if( geteuid() == 0 && setreuid( Daemon_uid, Daemon_uid ) < 0 ){
			logerr_die( XLOG_NOTICE, "Print_of_fd: setreuid failed" );
		}
		mexecv( cmd );
		logerr_die( XLOG_NOTICE,"Print_of_fd: execv failed %s", cmd);
	} else if( of_pid < 0 ){
		logerr_die( XLOG_NOTICE, "Print_of_fd: fork failed" );
	}
	(void)close(p[0]);		/* close input side */
	of_fd = p[1];
	of_running = of_pid;
	if(Debug>3)log(XLOG_DEBUG,"started OF process %d, fd %d", of_pid, of_fd);
}

/***************************************************************************
 * Print_close()
 * 1. Close the pipes and outputs.
 * 2. Signal the of process that it is to do something.
 *	Note that there is a Reapchild() call that will collect the
 *	children;  this will make sure that there is a minimal number of
 *	outstanding processes active.
 ***************************************************************************/

void
Print_close()
{
	if( of_fd ){
		(void)close(of_fd);
		of_fd = 0;
	}
	if( Print_fd ){
#ifdef	SYSV
		ioctl(Print_fd, TCFLSH, 1);
#endif
		if( RA ){
			Annex_ack();
		}
		(void)close(Print_fd);	/* close Printer */
		Print_fd = 0;
	}
	/*
	 * when you kill the printer, make sure that you do it neatly
	 */
	if( of_pid ){
#ifdef SIGCONT
		(void)kill(of_pid, SIGCONT);
#else
		(void)kill(of_pid, SIGUSR2);
#endif
	}
	of_pid = 0;
}


/***************************************************************************
 * Annex_ack()
 * 1. send an OOB ACK to indicate that you really have finished (EOF)
 * 2. Write a 0 byte to flush the output message (4.2 Bugger)
 * 3. Wait for a response that it has been done
 ***************************************************************************/
Annex_ack()
{
	int n;
	char nullstr[2];
#define ACK_CL_WAIT (10*60)

	(void)signal(SIGALRM, noannexack); /* set up time out function */
	if(setjmp(ack_jmp_buf)) {
		logerr( XLOG_INFO,
			"Annex_ack: '%s', port '%d' closing ack timed out",
			RA, AL );
	} else {
		nullstr[0] = 0;
		(void) alarm(ACK_CL_WAIT);
		if(Debug>3)log(XLOG_DEBUG, "Annex_ack: sending OOB");
		if (send(Print_fd, nullstr, 1, MSG_OOB) != 1) {
			(void) alarm(0); /* disarm */
			logerr(XLOG_INFO,"Annex_ack: sendoob failed" );
			return;
		}
		if(Debug>3)log(XLOG_DEBUG, "Annex_ack: sending null");
		if (send(Print_fd, nullstr, 1, 0) != 1) {
			(void) alarm(0); /* disarm */
			logerr(XLOG_INFO,"Annex_ack: send failed" );
			return;
		}
		if(Debug>3)log(XLOG_DEBUG, "Annex_ack: waiting for response");
		n = read( Print_fd, nullstr, 1 );
		(void) alarm(0); /* disarm */
		if( n != 1 ){
			logerr( XLOG_INFO, "Annex_ack: read error '%s', port '%d'",RA,AL);
		} else if( nullstr[0] != '\0' ){
			log( XLOG_INFO,
				"Annex_ack: '%d' read from '%s', port '%d'", nullstr[0],RA, AL);
		}
	}
}

/***************************************************************************
 * Print_ready()
 * 1. open the Printer if neccessary
 * 2. start up the output filter
 * 3. if the filter has been stopped, start it
 * Return: JSUCC if started, JFAIL if not
 ***************************************************************************/
Print_ready()
{
	/*
	 * open the Printer
	 */
	Print_open();
	Print_of_fd();
	if( of_pid && of_running == 0 ){
#ifdef SIGCONT
		if (kill(of_pid, SIGCONT) < 0) {
#else
		if (kill(of_pid, SIGUSR2) < 0) {
#endif			logerr( XLOG_INFO,"cannot restart output filter");
			Print_close();
			return( JFAIL );
		}
		of_running = of_pid;
	}
	return( JSUCC );
}

/***************************************************************************
 * of_stop()
 * stop the output filter if neccessary
 * 1. open and create the filter
 * 2. flush output
 * 3. kill(SIGSTOP) it
 * Return: JSUCC if stopped, JFAIL if not
 ***************************************************************************/

#ifdef	SYSV
int	sigusr1;
int	sigalrm;


void	handle_usr1()
{
  sigusr1++;
}

void	handle_alrm()
{
  sigalrm++;
}
#endif

int
of_stop()
{
	union wait statb;
	int pid;
#ifdef	SYSV
	void (*oldalrm)();
	void (*oldusr1)();
#endif

	Print_open();
	Print_of_fd();
	/*
	 * stop the OF filter
	 */
	if( of_pid && of_running ){
		if(Debug>3)log( XLOG_DEBUG, "stopping output filter" );
#ifdef	SYSV
		sigusr1 = sigalrm = 0;
		oldalrm = signal(SIGALRM, handle_alrm);
		oldusr1 = signal(SIGUSR1, handle_usr1);
#endif
		if( write( of_fd, filter_stop, strlen(filter_stop))
			!= strlen(filter_stop) ){
			logerr( XLOG_NOTICE,"of_stop: cannot write to OF");
			Print_close();
			return( JFAIL );
		}
#ifdef	SYSV
		/* 
		 * wait until OF sends SIGUSR1
		 */
		if(Debug>3)log( XLOG_DEBUG, "of_stop: waiting for OF %d", of_pid );
		
		alarm(5);	/* setup a alarm, if we lost a signal 
				   or the OF filter */
		while (!sigusr1 && !sigalrm) {
		  pause();
		}
		signal(SIGALRM, oldalrm);
		signal(SIGUSR1, oldusr1);
		if (sigalrm){
			logerr( XLOG_INFO, "of_stop: OF %d wont ACK STOP msg",
			       of_pid);
			Print_close();
			return( JFAIL );
		}
#else
		/*
		 * wait until OF blocks
		 */
		if(Debug>3)log( XLOG_DEBUG, "of_stop: waiting for OF %d", of_pid );
		while ((pid = wait3(&statb,WUNTRACED,(struct rusage *)0)) != -1
			&& pid != of_pid) ;
		if( pid < 0 || statb.w_stopval != WSTOPPED ){
			logerr( XLOG_INFO, "of_stop: OF %d died (%s)", of_pid,
				Decode_status( &statb ) );
			Print_close();
			return( JFAIL );
		}
#endif
		if(Debug>3)log( XLOG_DEBUG, "of_stop: output filter stopped" );
		of_running = 0;
	}
	return( JSUCC );
}

/***************************************************************************
 * Print_string( char *str )
 * print a string through a filter 
 * 1. Enable the line Printer
 * 2. get the filter or device fd;
 * 3. put out the string
 * 4. if unsuccessful, close the Printer
 * Return: JSUCC if successful, JFAIL otherwise
 ***************************************************************************/
int
Print_string( str )
	char *str;
{
	int f;
	int l;

	if( Print_ready() != JSUCC ){
		return( JFAIL );
	}
	if( of_fd ){
		f = of_fd;
	} else {
		f = Print_fd;
	}
	l = strlen( str );
	if( write( f, str, l ) != l ){
		logerr( XLOG_INFO, "Print_string: write error");
		Print_close();
		return( JFAIL );
	}
	return( JSUCC );
}


/***************************************************************************
 * Print_copy( int fd )
 * copy a file through a filter
 * 1. ready the Printer
 * 2. copy the file to the appropriate output device
 * 3. if an error, close the Printer
 ***************************************************************************/
int
Print_copy( fd )
	int fd;
{
	int f;
	long cnt;
	char buf[BUFSIZ];
	int in;			/* bytes read */
	int out;		/* bytes written out */

	cnt = 0;
	if( Print_ready() != JSUCC ){
		return( JFAIL );
	}
	if( of_fd ){
		f = of_fd;
	} else {
		f = Print_fd;
	}
	while( (in = read( fd, buf, sizeof(buf) )) > 0 ){
		out = write( f, buf, in );
		if( in != out ){
			logerr( XLOG_INFO, "Print_copy: write error");
			Print_close();
			return( JFAIL );
		}
		cnt = cnt + out;
	}
	/*
	 * completed the reading
	 */
	if( in < 0 ){
		logerr( XLOG_INFO, "Print_copy: read error");
		Print_close();
		return( JFAIL );
	}
	if(Debug>3)log(XLOG_DEBUG,"Print_copy: printed %d bytes", cnt);
	return( JSUCC );
}

/***************************************************************************
 *Print_filter( int fd, char *cmd )
 *  spawn a subprocess to do the printing
 * 1. stop the Printer
 * 2. fork a process
 * 3. wait for process to complete
 * 4. restart the process
 * Return: JSUCC if successful, JFAIL otherwise
 ***************************************************************************/
int
Print_filter( file, cmd )
	int file;
	char *cmd;
{
	int succ;			/* success code */
	int filter_pid;		/* filter process */
	int pid;			/* daughter pid */
	union wait status;	/* daughter status */
	int err;			/* saved value of errno */
	
	/*
	 * stop the Printer
	 */
	succ = of_stop();
	if( succ != JSUCC ){
		return( succ );
	}
	/*
	 * fork a process,  and connect file to fd 0, Printer to fd 1.
	 */
	if ((filter_pid = fork()) == 0) {	/* daughter */
		/*
		 * dup input file to standard in and Printer to stdout
		 */
		if( dup2(file, 0)<0 || dup2(Print_fd, 1)<0 ){
			logerr_die(XLOG_NOTICE,"Print_filter: dup2 failed filter %s",cmd);
		}
		mexecv(cmd);
		logerr_die( XLOG_NOTICE,"Print_filter: cannot execv %s", cmd);
	} else if( filter_pid < 0 ){
		logerr( XLOG_NOTICE, "Print_filter: fork failed" );
		Print_close();
		return( JFAIL );
	}
	/*
	 *	Wait for filter to complete 
	 */
	if(Debug>2)log(XLOG_DEBUG,"Print_filter: waiting for pid %d",filter_pid);
	while ((pid = wait(&status)) > 0 && pid != filter_pid){
		if(Debug>3)log( XLOG_DEBUG, "Print_filter:caught %d (%s)",
			pid, Decode_status(&status) );
	}
	err = errno;
	if(Debug>2)log( XLOG_DEBUG, "Print_filter: filter %d finished (%s)",
		pid,Decode_status(&status));
	errno = err;
	/*
	 * Check to see how filter terminated
	 */
	if( pid < 0 || !WIFEXITED(status)
		|| (unsigned)status.w_retcode > 1 ){
		/*
		 * died for bad reasons, don't run this again
		 */
		log(XLOG_INFO,"Print_filter:Filter '%s' Malfunction (%s)",
			cmd, Decode_status(&status));
		Print_close();
		return( JABORT );
	} else if (status.w_retcode != 0){
		/*
		 * try again
		 */
		log( XLOG_INFO, "Print_filter:Filter '%s' Retry wanted", cmd );
		Print_close();
		return(JFAIL);
	}
	return(JSUCC);
}

/*
 * Print_banner()
 * 1. get the Printer ready
 * 2. call the banner() routine with the correct parameter
 */
Print_banner()
{
	int f;

	if(Debug>3)log(XLOG_DEBUG,"Print_banner: printing banner");
	if( Print_ready() != JSUCC ){
		return( JFAIL );
	}
	if( of_fd ){
		f = of_fd;
	} else {
		f = Print_fd;
	}
	if( banner(f) != JSUCC ){
		if(Debug>3)log(XLOG_DEBUG,"Print_banner: banner failed");
		Print_close();
		return( JFAIL );
	}
	return( JSUCC );
}

/*
 * Fri Feb 26 08:44:53 CST 1988 Patrick Powell
 * set terminal modes
 * This code was based on a public domain version of public domain version
 * of stty. I suppose that I could have created if from scratermctrlh,
 * but I have seen the same table appearing
 * in many "public domain" display terminal modes programs.
 */
#ifdef	USE_TERMIO
struct termio tio;

struct tiomodes
{
	char	*string;
	int	set;
	int	reset;
};

struct tiomodes imodes[] = {
	{ "ignbrk",	IGNBRK,	IGNBRK	},
	{ "brkint",	BRKINT,	BRKINT	},
	{ "ignpar",	IGNPAR, IGNPAR	},
	{ "parmrk",	PARMRK,	PARMRK	},
	{ "inpck",	INPCK,	INPCK	},
	{ "istrip",	ISTRIP,	ISTRIP	},
	{ "inlcr",	INLCR,	INLCR	},
	{ "igncr",	IGNCR,	IGNCR	},
	{ "icrnl",	ICRNL,	ICRNL	},
	{ "iuclc",	IUCLC,	IUCLC	},
	{ "ixon",	IXON,	IXON	},
	{ "ixany",	IXANY,	IXANY	},
	{ "ixoff",	IXOFF,	IXOFF	},
	{ NULL,		0,	0 }
};

struct tiomodes omodes[] = {
	{ "opost",	OPOST,	OPOST	},
	{ "olcuc",	OLCUC,	OLCUC	},
	{ "onlcr",	ONLCR,	ONLCR	},
	{ "onocr",	ONOCR,	ONOCR	},
	{ "onlret",	ONLRET,	ONLRET	},
	{ "ofill",	OFILL,	OFILL	},
	{ "ofdel",	OFDEL,	OFDEL	},
	{ "nl0",	NL0,	NLDLY	},
	{ "nl1",	NL1,	NLDLY	},
	{ "cr0",	CR0,	CRDLY	},
	{ "cr1",	CR1,	CRDLY	},
	{ "cr2",	CR2,	CRDLY	},
	{ "cr3",	CR3,	CRDLY	},
	{ "tab0",	TAB0,	TABDLY	},
	{ "tab1",	TAB1,	TABDLY	},
	{ "tab2",	TAB2,	TABDLY	},
	{ "tab3",	TAB3,	TABDLY	},
	{ "bs0",	BS0,	BSDLY	},
	{ "bs1",	BS1,	BSDLY	},
	{ "vt0",	VT0,	VTDLY	},
	{ "vt1",	VT1,	VTDLY	},
	{ "ff0",	FF0,	FFDLY	},
	{ "ff1",	FF1,	FFDLY	},
	{ NULL,		0,	0	}
};
    
struct tiomodes cmodes[] = {
	{ "cs5",	CS5,	CSIZE	},
	{ "cs6",	CS6,	CSIZE	},
	{ "cs7",	CS7,	CSIZE	},
	{ "cs8",	CS8,	CSIZE	},
	{ "cstopb",	CSTOPB,	CSTOPB	},
	{ "cread",	CREAD,	CREAD	},
	{ "parenb",	PARENB,	PARENB	},
	{ "parodd",	PARODD,	PARODD	},
	{ "hupcl",	HUPCL,	HUPCL	},
	{ "clocal",	CLOCAL,	CLOCAL	},
	{ NULL,		0,	0	}
};

struct tiomodes lmodes[] = {
	{ "isig",	ISIG,	ISIG	},
	{ "icanon",	ICANON,	ICANON	},
	{ "xcase",	XCASE,	XCASE	},
	{ "echo",	ECHO,	ECHO	},
	{ "echoe",	ECHOE,	ECHOE	},
	{ "echok",	ECHOK,	ECHOK	},
	{ "echonl",	ECHONL,	ECHONL	},
	{ "noflsh",	NOFLSH,	NOFLSH	},
	{ NULL,		0,	0	}
};


Do_stty( fd )
	int fd;
{
	int i;
	int	localmode;
	int	linedisc;
	char buf[BUFSIZ], *bp, *ep, *arg;

	if( ioctl(fd, TCGETA, &tio) < 0) {
		logerr_die( XLOG_INFO,"cannot get tty parameters");
	}
	if(Debug>3)log(XLOG_DEBUG,"stty: before iflag 0x%x, oflag 0x%x, cflag 0x%x, lflag 0x%x",
			tio.c_iflag, tio.c_oflag, tio.c_cflag, tio.c_lflag);
	if(Debug>3) {
	  log(XLOG_DEBUG,"stty: pcap IS 0x%x, IC 0x%x, OS 0x%x, OC 0x%x,",
		IS, IC, OS, OC);
	  log(XLOG_DEBUG,"           CS 0x%x, CC 0x%x, LS 0x%x, LC 0x%x",
		CS, CC, LS, LC);
	}
	tio.c_iflag &= ~IC;
	tio.c_iflag |= IS;
	tio.c_oflag &= ~OC;
	tio.c_oflag |= OS;
	tio.c_cflag &= ~CC;
	tio.c_cflag |= CS;
	tio.c_lflag &= ~LC;
	tio.c_lflag |= LS;

	if( BR ){
	  for(i=0; bauds[i].baud && BR != bauds[i].baud; i++);
	  if( i == 0){
		  fatal(XLOG_INFO,"illegal baud rate %d", BR);
	  }
	  tio.c_cflag &= ~CBAUD;
	  tio.c_cflag |= bauds[i].speed;
	}
	if(Debug>3)log(XLOG_DEBUG,"stty: between iflag 0x%x, oflag 0x%x, cflag 0x%x, lflag 0x%x",
			tio.c_iflag, tio.c_oflag, tio.c_cflag, tio.c_lflag);

	if( TY && *TY ){
		(void)strcpy(buf, TY);
		ep = buf;
	} else {
		ep = 0;
	}
	while( ep && *ep ){
		for( ; *ep && isspace(*ep) ; ++ ep );
		for( arg = ep; *ep && !isspace(*ep) ; ++ ep );
		if( *ep ){
			*ep = 0;
			++ep;
		}
		for(i=0; imodes[i].string && strcmp(imodes[i].string,arg); i++);
		if(imodes[i].string) {
			if(Debug>4)log(XLOG_DEBUG,
				"stty: imodes %s, mc 0x%x ms 0x%x",
				imodes[i].string, imodes[i].reset, imodes[i].set);
			tio.c_iflag &= ~imodes[i].reset;
			tio.c_iflag |= imodes[i].set;
			continue;
		}

		for(i=0; omodes[i].string && strcmp(omodes[i].string,arg); i++);
		if(omodes[i].string) {
			if(Debug>4)log(XLOG_DEBUG,
				"stty: omodes %s, mc 0x%x ms 0x%x",
				omodes[i].string, omodes[i].reset, omodes[i].set);
			tio.c_oflag &= ~omodes[i].reset;
			tio.c_oflag |= omodes[i].set;
			continue;
		}

		for(i=0; cmodes[i].string && strcmp(cmodes[i].string,arg); i++);
		if(cmodes[i].string) {
			if(Debug>4)log(XLOG_DEBUG,
				"stty: cmodes %s, mc 0x%x ms 0x%x",
				cmodes[i].string, cmodes[i].reset, cmodes[i].set);
			tio.c_cflag &= ~cmodes[i].reset;
			tio.c_cflag |= cmodes[i].set;
			continue;
		}

		for(i=0; lmodes[i].string && strcmp(lmodes[i].string,arg); i++);
		if(lmodes[i].string) {
			if(Debug>4)log(XLOG_DEBUG,
				"stty: lmodes %s, mc 0x%x ms 0x%x",
				lmodes[i].string, lmodes[i].reset, lmodes[i].set);
			tio.c_lflag &= ~lmodes[i].reset;
			tio.c_lflag |= lmodes[i].set;
			continue;
		}
		for(i=0; bauds[i].string && strcmp(bauds[i].string,arg); i++);
		if(bauds[i].string) {
			if(Debug>4)log(XLOG_DEBUG,"stty: speed %s", arg );
			tio.c_cflag &= ~CBAUD;
			tio.c_cflag |= bauds[i].speed;
			continue;
		}
		fatal(XLOG_INFO,"unknown mode: %s\n", arg);
	}
	if(Debug>3)log(XLOG_DEBUG,"stty: after iflag 0x%x, oflag 0x%x, cflag 0x%x, lflag 0x%x",
			tio.c_iflag, tio.c_oflag, tio.c_cflag, tio.c_lflag);

	if( ioctl(fd, TCSETAW, &tio) < 0) {
		logerr_die( XLOG_NOTICE,"cannot set tty parameters");
	}
}
#else

struct tchars termctrl;
struct ltchars linectrl;
struct sgttyb mode;
struct
{
	char	*string;
	int	set;
	int	reset;
	int	lset;
	int	lreset;
} modes[] = {
	"bs0",		BS0, BS1, 0, 0,
	"bs1",		BS1, BS1, 0, 0,
	"cbreak",	CBREAK, 0, 0, 0,
	"-cbreak",	0, CBREAK, 0, 0,
	"cooked",	0, RAW, 0, 0,
	"cr0",		CR0, CR3, 0, 0,
	"cr1",		CR1, CR3, 0, 0,
	"cr2",		CR2, CR3, 0, 0,
	"cr3",		CR3, CR3, 0, 0,
	"decctlq",	0, 0, LDECCTQ, 0,
	"-decctlq",	0, 0, 0, LDECCTQ,
	"echo",		ECHO, 0, 0, 0,
	"-echo",	0, ECHO, 0, 0,
	"even",		EVENP, 0, 0, 0,
	"-even",	0, EVENP, 0, 0,
	"ff0",		FF0, FF1, 0, 0,
	"ff1",		FF1, FF1, 0, 0,
	"lcase",	LCASE, 0, 0, 0,
	"-lcase",	0, LCASE, 0, 0,
	"litout",	0, 0, LLITOUT, 0,
	"-litout",	0, 0, 0, LLITOUT,
	"nl",		0, CRMOD, 0, 0,
	"-nl",		CRMOD, 0, 0, 0,
	"nl0",		NL0, NL3, 0, 0,
	"nl1",		NL1, NL3, 0, 0,
	"nl2",		NL2, NL3, 0, 0,
	"nl3",		NL3, NL3, 0, 0,
	"noflsh",	0, 0, LNOFLSH, 0,
	"-noflsh",	0, 0, 0, LNOFLSH,
	"nohang",	0, 0, LNOHANG, 0,
	"-nohang",	0, 0, 0, LNOHANG,
	"odd",		ODDP, 0, 0, 0,
	"-odd",		0, ODDP, 0, 0,
	"raw",		RAW, 0, 0, 0,
	"-raw",		0, RAW, 0, 0,
	"tab0",		TAB0, XTABS, 0, 0,
	"tab1",		TAB1, XTABS, 0, 0,
	"tab2",		TAB2, XTABS, 0, 0,
	"tabs",		0, XTABS, 0, 0,
	"-tabs",	XTABS, 0, 0, 0,
	"tandem",	TANDEM, 0, 0, 0,
	"-tandem",	0, TANDEM, 0, 0,
	"tilde",	0, 0, LTILDE, 0,
	"-tilde",	0, 0, 0, LTILDE,
	"tn300",	CR1, ALLDELAY, 0, 0,
	"tty33",	CR1, ALLDELAY, 0, 0,
	"tty37",	FF1+CR2+TAB1+NL1, ALLDELAY, 0, 0,
	"vt05",		NL2, ALLDELAY, 0, 0,
	0,
};


struct	special {
	char	*name;
	char	*cp;
	char	def;
} special[] = {
	"stop",		&termctrl.t_stopc,		CSTOP,
	"start",	&termctrl.t_startc,		CSTART,
	0
};

Do_stty( fd )
	int fd;
{
	int i;
	int	localmode;
	int	linedisc;
	char buf[BUFSIZ], *bp, *ep, *arg;

	if( ioctl(fd, TIOCGETP, &mode) < 0
	 || ioctl(fd, TIOCGETC, &termctrl) < 0
	 || ioctl(fd, TIOCLGET, &localmode) < 0
	 || ioctl(fd, TIOCGLTC, &linectrl) < 0 ){
		logerr_die( XLOG_INFO,"cannot get tty parameters");
	}
	if(Debug>3)log(XLOG_DEBUG,"stty: before mode 0%o, lmode 0%o, speed %d",
			mode.sg_flags, localmode, mode.sg_ispeed );
	if( BR ){
	  for(i=0; bauds[i].baud && BR != bauds[i].baud; i++);
	  if( i == 0){
		  fatal(XLOG_INFO,"illegal baud rate %d", BR);
	  }
	  mode.sg_ispeed = mode.sg_ospeed = bauds[i].speed;
	}
	if (Debug>3)log(XLOG_DEBUG,
		 "stty: FC = 0%o, FS = 0%o, XC = 0%o, XS = 0%o",
		 FC, FS, XC, XS);
	mode.sg_flags &= ~FC;
	mode.sg_flags |= FS;
	localmode &= ~XC;
	localmode |= XS;
	if (Debug>3)log(XLOG_DEBUG, "stty: localmode = 0%o, mode.sg_flags= 0%o",
		        localmode, mode.sg_flags);
 

	if( TY && *TY ){
		(void)strcpy(buf, TY);
		ep = buf;
	} else {
		ep = 0;
	}
	while( ep && *ep ){
		for( ; *ep && isspace(*ep) ; ++ ep );
		for( arg = ep; *ep && !isspace(*ep) ; ++ ep );
		if( *ep ){
			*ep = 0;
			++ep;
		}
		for(i=0; modes[i].string && strcmp(modes[i].string,arg); i++);
		if(modes[i].string) {
			if(Debug>4)log(XLOG_DEBUG,
				"stty: modes %s, mc 0x%x ms 0x%x lc 0x%x ls 0x%x",
				modes[i].string, modes[i].reset, modes[i].set,
				modes[i].lreset, modes[i].lset );
			mode.sg_flags &= ~modes[i].reset;
			mode.sg_flags |= modes[i].set;
			localmode &= ~modes[i].lreset;
			localmode |= modes[i].lset;
			continue;
		}
		for (i = 0; special[i].name && strcmp(special[i].name,arg); i++);
		if( special[i].name ){
			for( ; *ep && isspace(*ep) ; ++ ep );
			for( bp = ep; *ep && !isspace(*ep) ; ++ ep );
			if( *ep ){
				*ep = 0;
				++ep;
			}
			if( *bp == 0 ){
				fatal( XLOG_INFO, "stty: missing parameter for %s", arg );
			}
			if (bp[0] == '^'){
				if( bp[1] == '?' ){
					*special[i].cp = 0177;
				} else {
					*special[i].cp = 037 & bp[1];
				}
			} else {
				*special[i].cp = bp[0];
			}
			if(Debug>4)log(XLOG_DEBUG,"stty: special %s %s", arg, bp );
			continue;
		}
		for(i=0; bauds[i].string && strcmp(bauds[i].string,arg); i++);
		if(bauds[i].string) {
			if(Debug>4)log(XLOG_DEBUG,"stty: speed %s", arg );
			mode.sg_ispeed = mode.sg_ospeed = bauds[i].speed;
			continue;
		}
		if (!strcmp("new", arg)){
			if(Debug>4)log(XLOG_DEBUG,"stty: ldisc %s", arg );
			linedisc = NTTYDISC;
			if (ioctl(fd, TIOCSETD, &linedisc)<0)
				logerr_die(XLOG_INFO,"stty: TIOCSETD ioctl failed");
			continue;
		}
		if (!strcmp("old",arg)){
			if(Debug>4)log(XLOG_DEBUG,"stty: ldisc %s", arg );
			linedisc = 0;
			if (ioctl(fd, TIOCSETD, &linedisc)<0)
				logerr_die(XLOG_INFO,"stty: TIOCSETD ioctl failed");
			continue;
		}
		fatal(XLOG_INFO,"unknown mode: %s\n", arg);
	}
	if(Debug>3)log(XLOG_DEBUG,"stty: after mode 0x%x, lmode 0x%x, speed 0x%x",
			mode.sg_flags, localmode, mode.sg_ispeed );
	if( ioctl(fd, TIOCSETN, &mode) < 0
		|| ioctl(fd, TIOCSETC, &termctrl) < 0
		|| ioctl(fd, TIOCSLTC, &linectrl) < 0
		|| ioctl(fd, TIOCLSET, &localmode) < 0 ){
		logerr_die( XLOG_NOTICE,"cannot set tty parameters");
	}
}


#endif USE_TERMIO





